#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.object module

This module defines ObjectWidget related classes.
"""

from zope.interface import Interface, alsoProvides, implementer
from zope.lifecycleevent import Attributes, ObjectCreatedEvent, ObjectModifiedEvent
from zope.schema import getFieldsInOrder
from zope.schema.interfaces import IObject

from pyams_form.converter import BaseDataConverter
from pyams_form.error import MultipleErrors
from pyams_form.field import FieldWidgets, Fields
from pyams_form.interfaces import DISPLAY_MODE, IDataConverter, INPUT_MODE, IObjectFactory
from pyams_form.interfaces.form import IFormAware
from pyams_form.interfaces.widget import IFieldWidget, IObjectWidget, IWidget
from pyams_form.widget import Widget, apply_value_to_widget
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config
from pyams_utils.factory import get_interface_name
from pyams_utils.interfaces.form import IDataManager, NO_VALUE
from pyams_utils.registry import get_current_registry


__docformat__ = 'restructuredtext'


class ObjectWidget_NO_VALUE:  # pylint: disable=invalid-name
    """No-value object widget placeholder"""
    def __repr__(self):
        return '<ObjectWidget_NO_VALUE>'


ObjectWidget_NO_VALUE = ObjectWidget_NO_VALUE()  # pylint: disable=invalid-name


class ObjectWidgetValue(dict):
    """Object widget value mapping"""
    original_value = ObjectWidget_NO_VALUE  # will store the original object


@adapter_config(required=(IObject, IObjectWidget),
                provides=IDataConverter)
class ObjectConverter(BaseDataConverter):
    """Data converter for IObjectWidget."""

    def to_widget_value(self, value):
        """Just dispatch it."""
        if value is self.field.missing_value:
            return NO_VALUE

        retval = ObjectWidgetValue()
        retval.original_value = value

        registry = self.widget.request.registry
        for name, field in getFieldsInOrder(self.field.schema):
            dman = registry.getMultiAdapter((value, field), IDataManager)
            subv = dman.query()

            if subv is NO_VALUE:
                # look up default value
                subv = field.default
                # XXX: too many discriminators
                # adapter = zope.component.queryMultiAdapter(
                #    (context, self.request, self.view, field, widget),
                #    interfaces.IValue, name='default')
                # if adapter:
                #    value = adapter.get()

            widget = registry.getMultiAdapter((field, self.widget.request),
                                              IFieldWidget)
            if IFormAware.providedBy(self.widget):
                # form property required by objectwidget
                widget.form = self.widget.form
                alsoProvides(widget, IFormAware)
            converter = registry.getMultiAdapter((field, widget), IDataConverter)
            retval[name] = converter.to_widget_value(subv)

        return retval

    def adapted_obj(self, obj):
        """Widget adapter object getter"""
        return self.field.schema(obj)

    def to_field_value(self, value):
        """field value is an Object type, that provides field.schema"""
        if value is NO_VALUE:
            return self.field.missing_value

        # try to get the original object, or if there's no chance an empty one
        obj = self.widget.get_object(value)
        obj = self.adapted_obj(obj)

        names = []
        registry = self.widget.request.registry
        for name, field in getFieldsInOrder(self.field.schema):
            if not field.readonly:
                try:
                    newval_raw = value[name]
                except KeyError:
                    continue

                widget = registry.getMultiAdapter((field, self.widget.request), IFieldWidget)
                converter = registry.getMultiAdapter((field, widget), IDataConverter)
                newval = converter.to_field_value(newval_raw)

                dman = registry.getMultiAdapter((obj, field), IDataManager)
                oldval = dman.query()
                if (oldval != newval) or IObject.providedBy(field):
                    dman.set(newval)
                    names.append(name)

        if names:
            registry.notify(ObjectModifiedEvent(obj, Attributes(self.field.schema, *names)))

        return obj


@implementer(IObjectWidget)
class ObjectWidget(Widget):
    """Object widget class"""

    _mode = INPUT_MODE
    _value = NO_VALUE
    _updating = False

    prefix = ''
    fields = None
    widgets = None

    def create_object(self, value):
        """Create widget object"""
        # keep value passed, maybe some subclasses want it
        # value here is the raw extracted from the widget's subform
        # in the form of a dict key:fieldname, value:fieldvalue
        name = get_interface_name(self.field.schema)
        registry = self.request.registry
        creator = registry.queryMultiAdapter((self.context, self.request, self.form, self),
                                             IObjectFactory, name=name)
        if creator:
            obj = creator(value)
        else:
            # raise RuntimeError, that won't be swallowed
            raise RuntimeError("No IObjectFactory adapter registered for %s" % name)
        return obj

    def get_object(self, value):
        """Get widget object"""
        if value.original_value is ObjectWidget_NO_VALUE:
            # if the original_value did not survive the roundtrip
            if self.ignore_context:
                obj = self.create_object(value)
            else:
                # try to get the original object from the context.field_name
                registry = self.request.registry
                dman = registry.getMultiAdapter((self.context, self.field), IDataManager)
                try:
                    obj = dman.get()
                except (KeyError, AttributeError):
                    obj = self.create_object(value)
        else:
            # reuse the object that we got in toWidgetValue
            obj = value.original_value

        if obj is None or obj == self.field.missing_value:
            # if still None, create one, otherwise following will burp
            obj = self.create_object(value)

        return obj

    @property
    def mode(self):
        """This sets the subwidgets modes."""
        return self._mode

    @mode.setter
    def mode(self, mode):
        """Mode setter applies mode to all widgets"""
        self._mode = mode
        # ensure that we apply the new mode to the widgets
        if self.widgets:
            for widget in self.widgets.values():
                widget.mode = mode

    def setup_fields(self):
        """Setup subform fields"""
        self.fields = Fields(self.field.schema)

    def setup_widgets(self):
        """Setup subform widgets"""
        self.setup_fields()

        self.prefix = self.name
        self.widgets = FieldWidgets(self, self.request, None)
        self.widgets.mode = self.mode
        # very-very important! otherwise the update() tries to set
        # RAW values as field values
        self.widgets.ignore_context = True
        self.widgets.ignore_request = self.ignore_request
        self.widgets.update()

    def update_widgets(self, set_errors=True):  # pylint: disable=unused-argument
        """Update subform widgets"""
        if self.field is None:
            raise ValueError("%r .field is None, that's a blocking point" % self)

        self.setup_widgets()

        if self._value is NO_VALUE:
            # XXX: maybe readonly fields/widgets should be reset here to
            #      widget.mode = INPUT_MODE
            for name, widget in self.widgets.items():
                if widget.field.readonly:
                    widget.mode = INPUT_MODE
                    widget.update()
        else:
            rawvalue = None

            for name, widget in self.widgets.items():
                if widget.mode == DISPLAY_MODE:
                    if rawvalue is None:
                        # lazy evaluation
                        registry = self.request.registry
                        converter = registry.getMultiAdapter((self.field, self),
                                                             IDataConverter)
                        obj = self.get_object(self._value)
                        rawvalue = converter.to_widget_value(obj)

                    self.apply_value(widget, rawvalue[name])
                else:
                    try:
                        val = self._value[name]
                    except KeyError:
                        pass
                    else:
                        self.apply_value(widget, val)

    def apply_value(self, widget, value):
        """Validate and apply value to given widget.

        This method gets called on any ObjectWidget value change and is
        responsible for validating the given value and setup an error message.

        This is internal apply value and validation process is needed because
        nothing outside this widget does know something about our
        internal sub widgets.
        """
        apply_value_to_widget(self, widget, value)

    def update(self):
        """Update widget contents"""
        # very-very-nasty: skip raising exceptions in extract while we're updating
        self._updating = True
        try:
            super().update()
            # create the subwidgets and set their values
            self.update_widgets(set_errors=False)
        finally:
            self._updating = False

    @property
    def value(self):
        """Widget value getter"""
        # value (get) cannot raise an exception, then we return insane values
        try:
            self.set_errors = True
            return self.extract()
        except MultipleErrors:
            value = ObjectWidgetValue()
            if self._value is not NO_VALUE:
                # send back the original object
                value.original_value = self._value.original_value

            for name, widget in self.widgets.items():
                if widget.mode != DISPLAY_MODE:
                    value[name] = widget.value
            return value

    @value.setter
    def value(self, value):
        """Widget value setter"""
        # This invokes updateWidgets on any value change e.g. update/extract.
        if (not isinstance(value, ObjectWidgetValue)) and (value is not NO_VALUE):
            value = ObjectWidgetValue(value)
        self._value = value

        # create the subwidgets and set their values
        self.update_widgets()

    def extract_raw(self, set_errors=True):
        '''See interfaces.form.IForm'''
        self.widgets.set_errors = set_errors
        data, errors = self.widgets.extract_raw()
        value = ObjectWidgetValue()
        if self._value is not NO_VALUE:
            # send back the original object
            value.original_value = self._value.original_value
        value.update(data)
        return value, errors

    def extract(self, default=NO_VALUE):
        if self.name + '-empty-marker' in self.request.params:
            self.update_widgets(set_errors=False)

            # important: widget extract MUST return RAW values
            # just an extractData is WRONG here
            value, errors = self.extract_raw(set_errors=self.set_errors)

            if errors:
                # very-very-nasty: skip raising exceptions in extract
                # while we're updating -- that happens when the widget
                # is updated and update calls extract()
                if self._updating:
                    # don't rebind value, send back the original object
                    for name, widget in self.widgets.items():
                        if widget.mode != DISPLAY_MODE:
                            value[name] = widget.value
                    return value
                raise MultipleErrors(errors)
            return value
        return default


def make_dummy_object(iface):
    """make dummy objects providing a given interface to support
    discriminating on field.schema
    """

    if iface is not None:
        @implementer(iface)
        class DummyObject:
            """Dummy object creator"""
    else:
        @implementer(Interface)
        class DummyObject:
            """Dummy object creator"""

    dummy = DummyObject()
    return dummy


@adapter_config(required=(Interface, IFormLayer, Interface, IWidget),
                provides=IObjectFactory)
class FactoryAdapter:
    """Most basic-default object factory adapter"""

    factory = None

    def __init__(self, context, request, form, widget):
        self.context = context
        self.request = request
        self.form = form
        self.widget = widget

    def __call__(self, value):
        # value is the extracted data from the form
        obj = self.factory()  # pylint: disable=not-callable
        registry = self.request.registry
        registry.notify(ObjectCreatedEvent(obj))
        return obj


# XXX: Probably we should offer a register factory method which allows to
# use all discriminators e.g. context, request, form, widget as optional
# arguments...
def register_factory_adapter(for_, klass, registry=None):
    """register the basic object factory adapter for a given interface and class"""

    class temp(FactoryAdapter):  # pylint: disable=invalid-name
        """Factory adapter class"""
        factory = klass

    name = get_interface_name(for_)
    if registry is None:
        registry = get_current_registry()
    registry.registerAdapter(temp,
                             required=(None, IFormLayer, None, IObjectWidget),
                             provided=IObjectFactory,
                             name=name)



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.image module

This module provides form's image button widget.
"""

from zope.interface import implementer_only
from zope.schema.fieldproperty import FieldProperty

from pyams_form.browser.button import ButtonWidget
from pyams_form.browser.interfaces import IHTMLImageWidget
from pyams_form.interfaces.button import IImageButton
from pyams_form.interfaces.widget import IFieldWidget, IImageWidget
from pyams_form.util import to_unicode
from pyams_form.widget import FieldWidget
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.form import NO_VALUE


__docformat__ = 'restructuredtext'


@implementer_only(IImageWidget)
class ImageWidget(ButtonWidget):
    """Form image button widget."""

    src = FieldProperty(IHTMLImageWidget['src'])

    klass = 'image-widget'
    css = 'image'

    def extract(self, default=NO_VALUE):
        """See pyams_form.interfaces.IWidget."""
        params = self.request.params
        if self.name + '.x' not in params:
            return default
        return {
            'x': int(params[self.name + '.x']),
            'y': int(params[self.name + '.y']),
            'value': params[self.name]
        }

    def json_data(self):
        data = super().json_data()
        data['type'] = 'image'
        data['src'] = self.src
        return data


@adapter_config(required=(IImageButton, IFormLayer), provides=IFieldWidget)
def ImageFieldWidget(field, request):  # pylint: disable=invalid-name
    """Form image button widget factory adapter"""
    image = FieldWidget(field, ImageWidget(request))
    image.value = field.title
    # Get the full resource URL for the image:
    image.src = to_unicode(field.image)
    return image



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.orderedselect module

Ordered-selection widget implementation.
"""

from zope.interface import implementer_only
from zope.schema.interfaces import IChoice, IList, ISequence, ITitledTokenizedTerm, ITuple

from pyams_form.browser.widget import HTMLSelectWidget, add_field_class
from pyams_form.interfaces.widget import IFieldWidget, IOrderedSelectWidget
from pyams_form.widget import FieldWidget, SequenceWidget


__docformat__ = "reStructuredText"

from pyams_layer.interfaces import IFormLayer

from pyams_utils.adapter import adapter_config


@implementer_only(IOrderedSelectWidget)
class OrderedSelectWidget(HTMLSelectWidget, SequenceWidget):
    """Ordered-Select widget implementation."""

    size = 5
    multiple = u'multiple'
    items = ()
    selected_items = ()
    notselected_items = ()

    def get_item(self, term, count=0):
        """Get item matching given term"""
        item_id = '%s-%i' % (self.id, count)
        content = term.value
        if ITitledTokenizedTerm.providedBy(term):
            content = self.request.localizer.translate(term.title)
        return {
            'id': item_id,
            'value': term.token,
            'content': content
        }

    def update(self):
        """See pyams_form.interfaces.widget.IWidget."""
        super().update()
        add_field_class(self)
        self.items = [
            self.get_item(term, count)
            for count, term in enumerate(self.terms)
        ]
        self.selected_items = list(filter(lambda x: x['value'] in self.value,
                                          self.items))
        self.notselected_items = list(filter(lambda x: x['value'] not in self.value,
                                             self.items))

    def json_data(self):
        """Get widget data in JSON format"""
        data = super().json_data()
        data['type'] = 'multi_select'
        data['options'] = self.items
        data['selected'] = self.selected_items
        data['not_selected'] = self.notselected_items
        return data


def OrderedSelectFieldWidget(field, request):  # pylint: disable=invalid-name
    """IFieldWidget factory for SelectWidget."""
    return FieldWidget(field, OrderedSelectWidget(request))


@adapter_config(required=(ISequence, IFormLayer), provided=IFieldWidget)
def SequenceSelectFieldWidget(field, request):  # pylint: disable=invalid-name
    """IFieldWidget factory for SelectWidget."""
    return request.registry.getMultiAdapter((field, field.value_type, request), IFieldWidget)


@adapter_config(required=(IList, IChoice, IFormLayer), provided=IFieldWidget)
@adapter_config(required=(ITuple, IChoice, IFormLayer), provided=IFieldWidget)
def SequenceChoiceSelectFieldWidget(field, value_type, request):
    # pylint: disable=invalid-name,unused-argument
    """IFieldWidget factory for SelectWidget."""
    return OrderedSelectFieldWidget(field, request)



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.datamanager module

This module provides base data manager classes.
"""

from zope.interface import implementer, Interface
from zope.interface.common.mapping import IMapping
from zope.schema.interfaces import IField
from zope.security import canAccess, canWrite
from zope.security.interfaces import ForbiddenAttribute
from zope.security.proxy import Proxy

from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.form import IDataManager, NO_VALUE


__docformat__ = 'restructuredtext'


_MARKER = []


ALLOWED_DATA_CLASSES = [dict]
try:
    import persistent.mapping
    import persistent.dict
    ALLOWED_DATA_CLASSES.append(persistent.mapping.PersistentMapping)
    ALLOWED_DATA_CLASSES.append(persistent.dict.PersistentDict)
except ImportError:
    pass


@implementer(IDataManager)
class DataManager:
    """Data manager base class."""


@adapter_config(required=(Interface, IField), provides=IDataManager)
class AttributeField(DataManager):
    """Attribute field."""

    def __init__(self, context, field):
        self.context = context
        self.field = field

    @property
    def adapted_context(self):
        """Data manager adapted context getter"""
        # get the right adapter or context
        context = self.context
        # NOTE: zope.schema fields defined in inherited interfaces will point
        # to the inherited interface. This could end in adapting the wrong item.
        # This is very bad because the widget field offers an explicit interface
        # argument which doesn't get used in Widget setup during IDataManager
        # lookup. We should find a concept which allows to adapt the
        # IDataManager use the widget field interface instead of the zope.schema
        # field.interface, ri
        if self.field.interface is not None:
            context = self.field.interface(context)
        return context

    def get(self):
        """See pyams_utils.interfaces.form.IDataManager"""
        return getattr(self.adapted_context, self.field.__name__)

    def query(self, default=NO_VALUE):
        """See pyams_utils.interfaces.form.IDataManager"""
        try:
            return self.get()
        except ForbiddenAttribute as e:  # pylint: disable=invalid-name
            raise e
        except AttributeError:
            return default

    def set(self, value):
        """See pyams_utils.interfaces.form.IDataManager"""
        if self.field.readonly:
            raise TypeError("Can't set values on read-only fields (name=%s, class=%s.%s)" %
                            (self.field.__name__,
                             self.context.__class__.__module__,
                             self.context.__class__.__name__))
        # get the right adapter or context
        setattr(self.adapted_context, self.field.__name__, value)

    def can_access(self):
        """See pyams_utils.interfaces.form.IDataManager"""
        context = self.adapted_context
        if isinstance(context, Proxy):
            return canAccess(context, self.field.__name__)
        return True

    def can_write(self):
        """See pyams_utils.interfaces.form.IDataManager"""
        context = self.adapted_context
        if isinstance(context, Proxy):
            return canWrite(context, self.field.__name__)
        return True


@adapter_config(required=(dict, IField), provides=IDataManager)
class DictionaryField(DataManager):
    """Dictionary field.

    NOTE: Even though, this data manager allows nearly all kinds of
    mappings, by default it is only registered for dict, because it
    would otherwise get picked up in undesired scenarios. If you want
    to it use for another mapping, register the appropriate adapter in
    your application.

    """

    _allowed_data_classes = tuple(ALLOWED_DATA_CLASSES)

    def __init__(self, data, field):
        if (not isinstance(data, self._allowed_data_classes) and
                not IMapping.providedBy(data)):
            raise ValueError("Data are not a dictionary: %s" % type(data))
        self.data = data
        self.field = field

    def get(self):
        """See pyams_utils.interfaces.form.IDataManager"""
        value = self.data.get(self.field.__name__, _MARKER)
        if value is _MARKER:
            raise AttributeError
        return value

    def query(self, default=NO_VALUE):
        """See pyams_utils.interfaces.form.IDataManager"""
        return self.data.get(self.field.__name__, default)

    def set(self, value):
        """See pyams_utils.interfaces.form.IDataManager"""
        if self.field.readonly:
            raise TypeError("Can't set values on read-only fields name=%s"
                            % self.field.__name__)
        self.data[self.field.__name__] = value

    @staticmethod
    def can_access():
        """See pyams_utils.interfaces.form.IDataManager"""
        return True

    @staticmethod
    def can_write():
        """See pyams_utils.interfaces.form.IDataManager"""
        return True



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.multi module

This module provides multi-widgets implementation.
"""

from operator import attrgetter

from zope.interface import implementer
from zope.schema.interfaces import IDict, IField, IList, ITuple

from pyams_form.browser.widget import HTMLFormElement
from pyams_form.button import Buttons, button_and_handler
from pyams_form.interfaces.button import IActions
from pyams_form.interfaces.form import IButtonForm, IHandlerForm
from pyams_form.interfaces.widget import IFieldWidget, IMultiWidget
from pyams_form.widget import FieldWidget, MultiWidget as MultiWidgetBase
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config


__docformat__ = 'restructuredtext'

from pyams_form import _  # pylint: disable=ungrouped-imports


@implementer(IButtonForm, IHandlerForm)
class FormMixin:
    """Form mixin class"""


@implementer(IMultiWidget)
class MultiWidget(HTMLFormElement, MultiWidgetBase, FormMixin):
    # pylint: disable=function-redefined
    """Multi widget implementation."""

    buttons = Buttons()

    prefix = 'widget'
    klass = 'multi-widget'
    css = 'multi'
    items = ()

    actions = None
    show_label = True  # show labels for item subwidgets or not

    # Internal attributes
    _adapter_value_attributes = MultiWidgetBase._adapter_value_attributes + ('show_label',)

    def update(self):
        """See pyams_form.interfaces.widget.IWidget."""
        super().update()
        self.update_actions()
        self.actions.execute()
        self.update_actions()  # Update again, as conditions may change

    def update_actions(self):
        """Update widget actions"""
        self.update_allow_add_remove()
        if self.name is not None:
            self.prefix = self.name
        registry = self.request.registry
        self.actions = registry.getMultiAdapter((self, self.request, self), IActions)
        self.actions.update()

    @button_and_handler(_('Add'), name='add',
                        condition=attrgetter('allow_adding'))
    def handle_add(self, action):  # pylint: disable=unused-argument
        """Add button handler"""
        self.append_adding_widget()

    @button_and_handler(_('Remove selected'), name='remove',
                        condition=attrgetter('allow_removing'))
    def handle_remove(self, action):  # pylint: disable=unused-argument
        """Remove button handler"""
        self.remove_widgets([widget.name for widget in self.widgets
                             if '{}.remove'.format(widget.name) in self.request.params])


@adapter_config(required=(IDict, IFormLayer),
                provided=IFieldWidget)
def MultiFieldWidgetFactory(field, request):  # pylint: disable=invalid-name
    """IFieldWidget factory for MultiWidget."""
    return FieldWidget(field, MultiWidget(request))


@adapter_config(required=(IDict, IField, IFormLayer),
                provided=IFieldWidget)
@adapter_config(required=(IList, IField, IFormLayer),
                provided=IFieldWidget)
@adapter_config(required=(ITuple, IField, IFormLayer),
                provided=IFieldWidget)
def MultiFieldWidget(field, value_type, request):  # pylint: disable=invalid-name,unused-argument
    """IFieldWidget factory for MultiWidget."""
    return MultiFieldWidgetFactory(field, request)



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.interfaces.form module

Form related interfaces module.
"""

from zope.interface import Attribute, Interface
from zope.interface.interfaces import IObjectEvent
from zope.lifecycleevent import IObjectCreatedEvent
from zope.schema import ASCIILine, Bool, Choice, Dict, Field, Object, Text, TextLine, Tuple, URI

from pyams_form.interfaces import DISPLAY_MODE, IContentProviders, IFields, INPUT_MODE
from pyams_form.interfaces.button import IActions, IButtonHandlers, IButtons
from pyams_form.interfaces.widget import IWidgets


__docformat__ = 'restructuredtext'

from pyams_form import _


class IHandlerForm(Interface):
    """A form that stores the handlers locally."""

    handlers = Object(title=_('Handlers'),
                      description=_('A list of action handlers defined on the form.'),
                      schema=IButtonHandlers,
                      required=True)


class IActionForm(Interface):
    """A form that stores executable actions"""

    actions = Object(title=_('Actions'),
                     description=_('A list of actions defined on the form'),
                     schema=IActions,
                     required=True)

    refresh_actions = Bool(title=_('Refresh actions'),
                           description=_('A flag, when set, causes form actions to be '
                                         'updated again after their execution.'),
                           default=False,
                           required=True)


class IContextAware(Interface):
    """Offers a context attribute.

    For advanced uses, the widget will make decisions based on the context
    it is rendered in.
    """

    context = Field(title=_('Context'),
                    description=_('The context in which the widget is displayed.'),
                    required=True)

    ignore_context = Bool(title=_('Ignore Context'),
                          description=_('A flag, when set, forces the widget not to look at '
                                        'the context for a value.'),
                          default=False,
                          required=False)


class IFormAware(Interface):
    """Offers a form attribute.

    For advanced uses the widget will make decisions based on the form
    it is rendered in.
    """

    form = Field()


class IForm(Interface):
    """Form interface"""

    fields = Attribute("Form fields")

    mode = Choice(title=_('Mode'),
                  description=_('The mode in which to render the widgets.'),
                  values=(INPUT_MODE, DISPLAY_MODE),
                  required=True)

    ignore_context = Bool(title=_('Ignore Context'),
                          description=_('If set the context is ignored to retrieve a value.'),
                          default=False,
                          required=True)

    ignore_request = Bool(title=_('Ignore Request'),
                          description=_('If set the request is ignored to retrieve a value.'),
                          default=False,
                          required=True)

    ignore_readonly = Bool(title=_('Ignore Readonly'),
                           description=_('If set then readonly fields will also be shown.'),
                           default=False,
                           required=True)

    ignore_required_on_extract = Bool(
        title=_('Ignore Required validation on extract'),
        description=_("If set then required fields will pass validation "
                      "on extract regardless whether they're filled in or not"),
        default=False,
        required=True)

    widgets = Object(title=_('Widgets'),
                     description=_('A widget manager containing the widgets to be used in '
                                   'the form.'),
                     schema=IWidgets)

    title = TextLine(title=_('Title'),
                     description=_('Main form title'),
                     required=False)

    subtitle = Attribute("Form subtitle")

    legend = TextLine(title=_('Legend'),
                      description=_('A human readable text describing the form that can be '
                                    'used in the UI.'),
                      required=False)

    required_label = TextLine(title=_('Required label'),
                              description=_('A human readable text describing the form that can '
                                            'be used in the UI for rendering a required info '
                                            'legend.'),
                              required=False)

    prefix = ASCIILine(title=_('Prefix'),
                       description=_('The prefix of the form used to uniquely identify it.'),
                       default='form.')

    status = Text(title=_('Status'),
                  description=_('The status message of the form.'),
                  default=None,
                  required=False)

    def get_content(self):
        """Return the content to be displayed and/or edited."""

    def update_widgets(self, prefix=None):
        """Update the widgets for the form.

        This method is commonly called from the ``update()`` method and is
        mainly meant to be a hook for subclasses.

        Note that you can pass an argument for ``prefix`` to override
        the default value of ``"widgets."``.
        """

    def extract_data(self, set_errors=True, notify=True):
        """Extract the data of the form.

        set_errors: needs to be passed to extract() and to sub-widgets
        notify: set to False when you don't want to notify IDataExtractedEvent"""

    def update(self):
        """Update the form."""

    def render(self):
        """Render the form."""

    def json(self):
        """Returns the form in json format"""


class IAJAXForm(IForm):
    """A form interface for handling AJAX calls"""

    def get_ajax_handler(self):
        """Get absolute URL of AJAX handler"""

    def get_form_options(self):
        """Get form options in JSON format"""

    def get_ajax_errors(self):
        """Get AJAX status"""

    def get_ajax_output(self, changes):
        """Get AJAX POST output in JSON"""


class IAJAXFormRenderer(Interface):
    """AJAX form JSON renderer"""

    def render(self, changes):
        """Render changes in JSON"""


class ISubForm(IForm):
    """A subform."""

    parent_form = Attribute("Parent form")


class IInnerForm(Interface):
    """Inner form marker interface"""


class IInnerSubForm(IInnerForm, ISubForm):
    """Inner sub-form interface

    Inner sub-forms are standard sub-forms, but which may be included dynamically
    into a given form by using named adapters.
    """


class IInnerTabForm(IInnerForm, ISubForm):
    """Inner tab-form interface

    Inner tab-forms are standard sub-forms, but which may be included dynamically
    into a given form by using named adapters.
    """


class IDisplayForm(IForm):
    """Mark a form as display form, used for templates."""


class IInputForm(Interface):
    """A form that is meant to process the input of the form controls."""

    action = URI(title=_('Action'),
                 description=_('The action defines the URI to which the form data are '
                               'sent.'),
                 required=True)

    name = TextLine(title=_('Name'),
                    description=_('The name of the form used to identify it.'),
                    required=False)

    # pylint: disable=invalid-name
    id = TextLine(title=_('Id'),
                  description=_('The id of the form used to identify it.'),
                  required=False)

    method = Choice(title=_('Method'),
                    description=_('The HTTP method used to submit the form.'),
                    values=('get', 'post'),
                    default='post',
                    required=False)

    enctype = ASCIILine(title=_('Encoding Type'),
                        description=_('The data encoding used to submit the data safely.'),
                        default='multipart/form-data',
                        required=False)

    accept_charset = ASCIILine(title=_('Accepted Character Sets'),
                               description=_('This is a list of character sets the server '
                                             'accepts. By default this is unknown.'),
                               required=False)

    accept = ASCIILine(title=_('Accepted Content Types'),
                       description=_('This is a list of content types the server can '
                                     'safely handle.'),
                       required=False)

    autocomplete = Choice(title=_("Form autocomplete"),
                          description=_("Enable or disable global form autocomplete"),
                          values=('on', 'off', 'new-password'),
                          required=False)

    # AJAX related form settings

    ajax_form_handler = TextLine(title="Name of AJAX form handler",
                                 required=False)

    ajax_form_options = Dict(title="AJAX form submit's data options",
                             required=False)

    ajax_form_target = TextLine(title="Form submit target",
                                description="Form content target, used for HTML and text content "
                                            "types",
                                required=False)

    ajax_form_callback = TextLine(title="AJAX submit callback",
                                  description="Name of a custom form submit callback",
                                  required=False)

    def get_ajax_handler(self):
        """Get absolute URL of AJAX handler"""

    def get_form_options(self):
        """Get form options in JSON format"""


class IAddForm(IForm):
    """A form to create and add a new component."""

    content_factory = Attribute("Form new content factory")

    def create_and_add(self, data):
        """Call create and add.

        This method can be used for keep all attributes internal during create
        and add calls. On sucess we return the new created and added object.
        If something fails, we return None. The default handleAdd method will
        only set the _finished_add marker on sucess.
        """

    def create(self, data):
        """Create the new object using the given data.

        Returns the newly created object.
        """

    def add(self, obj):
        """Add the object somewhere."""

    def update_content(self, obj, data):
        """Update content after creation

        This is actually used to apply updates in subforms to newly created content.
        """


class IEditForm(IForm):
    """A form to edit data of a component."""

    def apply_changes(self, data):
        """Apply the changes to the content component."""


class IFieldsForm(IForm):
    """A form that is based upon defined fields."""

    fields = Object(title=_('Fields'),
                    description=_('A field manager describing the fields to be used for '
                                  'the form.'),
                    schema=IFields)


class IFieldsAndContentProvidersForm(IForm):
    """A form that is based upon defined fields and content providers"""

    content_providers = Object(title=_('Content providers'),
                               description=_(
                                   'A manager describing the content providers to be used for '
                                   'the form.'),
                               schema=IContentProviders)


class IButtonForm(IForm):
    """A form that is based upon defined buttons."""

    buttons = Object(title=_('Buttons'),
                     description=_('A button manager describing the buttons to be used for '
                                   'the form.'),
                     schema=IButtons)


class IGroupManager(IForm):
    """Groups manager interface"""

    groups = Tuple(title='Groups',
                   description=('Initially a collection of group classes, which are '
                                'converted to group instances when the form is '
                                'updated.'))

    def get_groups(self):
        """Internal groups iterator getter"""


class IGroup(IGroupManager):
    """A group of fields/widgets within a form."""

    parent_form = Attribute("Parent form")


class IGroupForm(IGroupManager):
    """A form that supports groups."""


class IFormCreatedEvent(IObjectCreatedEvent):
    """Form created event interface"""

    request = Attribute("Form request")


class IFormUpdatedEvent(IObjectEvent):
    """Form updated event interface"""

    form = Attribute("Form")
    request = Attribute("Form request")


class IFormContent(Interface):
    """Optional adapter interface used to get form content"""


class IFormFields(Interface):
    """Optional adapter interface used to get form fields"""


class IDataExtractedEvent(Interface):
    """Event sent after data and status are extracted from widgets.
    """
    data = Attribute("Extracted form data. Usually, the widgets extract field names from "
                     "the request and return a dictionary of field names and field values.")
    errors = Attribute("Tuple of status providing IErrorViewSnippet.")
    form = Attribute("Form instance.")



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.textarea module

This module provides textarea widget.
"""

from zope.interface import implementer_only
from zope.schema.interfaces import IASCII, IText

from pyams_form.browser.widget import HTMLTextAreaWidget, add_field_class
from pyams_form.interfaces.widget import IFieldWidget, ITextAreaWidget
from pyams_form.widget import FieldWidget, Widget
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config


__docformat__ = 'restructuredtext'


@implementer_only(ITextAreaWidget)
class TextAreaWidget(HTMLTextAreaWidget, Widget):
    """Textarea widget implementation."""

    klass = 'textarea-widget'
    css = 'textarea'
    value = ''

    def update(self):
        super().update()
        add_field_class(self)

    def json_data(self):
        data = super().json_data()
        data['type'] = 'textarea'
        return data


@adapter_config(required=(IASCII, IFormLayer),
                provides=IFieldWidget)
@adapter_config(required=(IText, IFormLayer),
                provides=IFieldWidget)
def TextAreaFieldWidget(field, request):  # pylint: disable=invalid-name
    """IFieldWidget factory for TextWidget."""
    return FieldWidget(field, TextAreaWidget(request))



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.testing module

This module provides several testing helpers.
"""

import base64
import os
import pprint
import re
import sys
from doctest import register_optionflag

import lxml
from pyramid.testing import DummyRequest
from zope.component import adapts
from zope.i18n.locales import locales
from zope.interface import Interface, implementer, provider
from zope.schema import Bool, Choice, Date, Dict, Int, List, Object, TextLine
from zope.schema.fieldproperty import FieldProperty
from zope.schema.interfaces import IBytes
from zope.security import checker
from zope.security.interfaces import IInteraction, ISecurityPolicy

from pyams_form import browser, field, outputchecker
from pyams_form.ajax import ajax_form_config
from pyams_form.converter import FileUploadDataConverter
from pyams_form.form import AddForm, EditForm
from pyams_form.interfaces.form import IAJAXForm
from pyams_form.interfaces.widget import IFileWidget
from pyams_layer.interfaces import ISkin, PYAMS_BASE_SKIN_NAME
from pyams_layer.skin import PyAMSSkin, apply_skin


__docformat__ = 'restructuredtext'


if sys.argv[-1].endswith('/bin/test'):

    lxml.doctestcompare.NOPARSE_MARKUP = register_optionflag('NOPARSE_MARKUP')

    OUTPUT_CHECKER = outputchecker.OutputChecker(
        patterns=(
            (re.compile(r"u('.*?')"), r"\1"),
            (re.compile(r"b('.*?')"), r"\1"),
            (re.compile(r"__builtin__"), r"builtins"),
            (re.compile(r"<type"), r"<class"),
            (re.compile(r"set\(\[(.*?)\]\)"), r"{\1}"),
        )
    )

    INT_LABEL = 'Int label'
    BOOL_LABEL = 'Bool label'
    CHOICE_LABEL = 'Choice label'
    CHOICE_OPT_LABEL = 'ChoiceOpt label'
    TEXTLINE_LABEL = 'TextLine label'
    DATE_LABEL = 'Date label'
    READONLY_LABEL = 'ReadOnly label'
    OBJECT_LABEL = 'Object label'


    def TestRequest(**kwargs):  # pylint: disable=invalid-name
        """Test request helper"""
        params = kwargs.get('params')
        if params:
            copy = params.__class__()
            for key, value in params.items():
                if key.endswith(':list'):
                    key = key.split(':', 1)[0]
                copy[key] = value
            kwargs['params'] = copy
        request = DummyRequest(**kwargs)
        request.locale = locales.getLocale('en')
        apply_skin(request, PYAMS_BASE_SKIN_NAME)
        return request


    class TestingFileUploadDataConverter(FileUploadDataConverter):
        """A special file upload data converter that works with testing."""

        adapts(IBytes, IFileWidget)

        def to_field_value(self, value):
            if value is None or value == '':
                value = self.widget.request.params.get(self.widget.name + '.testing', '')
                encoding = self.widget.request.params.get(
                    self.widget.name + '.encoding', 'plain')

                # allow for the case where the file contents are base64 encoded.
                if encoding == 'base64':
                    value = base64.b64decode(value)
                self.widget.request.POST[self.widget.name] = value

            return super().to_field_value(value)


    @implementer(IInteraction)
    @provider(ISecurityPolicy)
    class SimpleSecurityPolicy:
        """Allow all access."""

        logged_in = False
        allowed_permissions = ()
        participations = None

        def __init__(self, logged_in=False, allowed_permissions=()):
            self.logged_in = logged_in
            self.allowed_permissions = allowed_permissions + (checker.CheckerPublic,)

        def __call__(self, *participations):
            self.participations = []
            return self

        def checkPermission(self, permission, object):
            # pylint: disable=invalid-name,unused-argument,redefined-builtin
            """Check given permission"""
            if self.logged_in:
                if permission in self.allowed_permissions:
                    return True
            return False


    def get_path(filename):
        """Get full path of given filename"""
        return os.path.join(os.path.dirname(browser.__file__), filename)


    def text_of_with_optional_title(node, add_title=False, show_tooltips=False):
        # pylint: disable=too-many-branches,too-many-return-statements
        """Get node text with title"""
        if isinstance(node, (list, tuple)):
            return '\n'.join(text_of_with_optional_title(child, add_title, show_tooltips)
                             for child in node)
        text = []
        if node is None:
            return None

        if node.tag == 'br':
            return '\n'
        if node.tag == 'input':
            if add_title:
                title = node.get('name') or ''
                title += ' '
            else:
                title = ''
            if node.get('type') == 'radio':
                return title + ('(O)' if node.get('checked') else '( )')
            if node.get('type') == 'checkbox':
                return title + ('[x]' if node.get('checked') else '[ ]')
            if node.get('type') == 'hidden':
                return ''
            return '%s[%s]' % (title, node.get('value') or '')
        if node.tag == 'textarea':
            if add_title:
                title = node.get('name') or ''
                title += ' '
                text.append(title)
        if node.tag == 'select':
            if add_title:
                title = node.get('name') or ''
                title += ' '
            else:
                title = ''
            option = node.find('option[@selected]')
            return '%s[%s]' % (title, option.text if option is not None else '[    ]')
        if node.tag == 'li':
            text.append('*')
        if node.tag == 'script':
            return None

        if node.text and node.text.strip():
            text.append(node.text.strip())

        for _index, child in enumerate(node):
            txt = text_of_with_optional_title(child, add_title, show_tooltips)
            if txt:
                text.append(txt)
            if child.tail and child.tail.strip():
                text.append(child.tail)
        text = ' '.join(text).strip()
        # 'foo<br>bar' ends up as 'foo \nbar' due to the algorithm used above
        text = text.replace(' \n', '\n').replace('\n ', '\n').replace('\n\n', '\n')
        if u'\xA0' in text:
            # don't just .replace, that'll sprinkle my tests with u''
            text = text.replace(u'\xA0', ' ')  # nbsp -> space
        if node.tag == 'li':
            text += '\n'
        if node.tag == 'div':
            text += '\n'
        return text


    def text_of(node):
        """Return the contents of an HTML node as text.

        Useful for functional tests, e.g. ::

            print map(textOf, browser.etree.xpath('.//td'))

        """
        return text_of_with_optional_title(node, False)


    def plain_text(content, xpath=None):
        """Get text of given node"""
        root = lxml.html.fromstring(content)
        if xpath is not None:
            nodes = root.xpath(xpath)
            joinon = '\n'
        else:
            nodes = root
            joinon = ''
        text = joinon.join(map(text_of, nodes))
        lines = [l.strip() for l in text.splitlines()]
        text = '\n'.join(lines)
        return text


    def get_submit_values(content):
        """Get submit values from content"""
        root = lxml.html.fromstring(content)
        form = root.forms[0]
        values = dict(form.form_values())
        return values


    #
    # classes required by ObjectWidget tests
    #

    class IMySubObject(Interface):
        """Sub-object interface"""
        foofield = Int(title="My foo field",
                       default=1111,
                       max=9999,
                       required=True)
        barfield = Int(title="My dear bar",
                       default=2222,
                       required=False)


    @implementer(IMySubObject)
    class MySubObject:
        """Sub-object class"""
        foofield = FieldProperty(IMySubObject['foofield'])
        barfield = FieldProperty(IMySubObject['barfield'])


    class IMySecond(Interface):
        """Second interface"""
        subfield = Object(title="Second-subobject",
                          schema=IMySubObject)
        moofield = TextLine(title="Something")


    @implementer(IMySecond)
    class MySecond:
        """Second object class"""
        subfield = FieldProperty(IMySecond['subfield'])
        moofield = FieldProperty(IMySecond['moofield'])


    class IMyObject(Interface):
        """Object interface"""
        subobject = Object(title='my object', schema=IMySubObject)
        name = TextLine(title='name')


    @implementer(IMyObject)
    class MyObject:
        """Object class"""

        def __init__(self, name='', subobject=None):
            self.subobject = subobject
            self.name = name


    class IMyComplexObject(Interface):
        """Complex object interface"""
        subobject = Object(title='my object', schema=IMySecond)
        name = TextLine(title='name')


    class IMySubObjectMulti(Interface):
        """Multi sub-object interface"""
        foofield = Int(title="My foo field",
                       default=None,  # default is None here!
                       max=9999,
                       required=True)
        barfield = Int(title="My dear bar",
                       default=2222,
                       required=False)


    @implementer(IMySubObjectMulti)
    class MySubObjectMulti:
        """Multi sub-object class"""
        foofield = FieldProperty(IMySubObjectMulti['foofield'])
        barfield = FieldProperty(IMySubObjectMulti['barfield'])


    class IMyMultiObject(Interface):
        """Multi-object interface"""
        list_of_objects = List(title="My list field",
                               value_type=Object(
                                   title='my object widget',
                                   schema=IMySubObjectMulti))
        name = TextLine(title='name')


    @implementer(IMyMultiObject)
    class MyMultiObject:
        """Multi-object class"""
        list_of_objects = FieldProperty(IMyMultiObject['list_of_objects'])
        name = FieldProperty(IMyMultiObject['name'])

        def __init__(self, name='', list_of_objects=None):
            self.list_of_objects = list_of_objects
            self.name = name


    class IntegrationBase:
        """Integration base"""
        def __init__(self, **kw):
            for key, value in kw.items():
                setattr(self, key, value)

        def __repr__(self):
            items = list(self.__dict__.items())
            items.sort()
            return ("<" + self.__class__.__name__+"\n  " + "\n  ".join(
                ["%s: %s" % (key, pprint.pformat(value)) for key, value in items]) + ">")


    class IObjectWidgetSingleSubIntegration(Interface):
        """Sub-integration interface"""
        single_int = Int(title=INT_LABEL)
        single_bool = Bool(title=BOOL_LABEL)
        single_choice = Choice(title=CHOICE_LABEL,
                               values=('one', 'two', 'three'))
        single_choice_opt = Choice(title=CHOICE_OPT_LABEL,
                                   values=('four', 'five', 'six'),
                                   required=False)
        single_textline = TextLine(title=TEXTLINE_LABEL)
        single_date = Date(title=DATE_LABEL)
        single_readonly = TextLine(title=READONLY_LABEL,
                                   readonly=True)


    @implementer(IObjectWidgetSingleSubIntegration)
    class ObjectWidgetSingleSubIntegration(IntegrationBase):
        """Sub-integration class"""
        single_int = FieldProperty(IObjectWidgetSingleSubIntegration['single_int'])
        single_bool = FieldProperty(IObjectWidgetSingleSubIntegration['single_bool'])
        single_choice = FieldProperty(IObjectWidgetSingleSubIntegration['single_choice'])
        single_choice_opt = FieldProperty(
            IObjectWidgetSingleSubIntegration['single_choice_opt'])
        single_textline = FieldProperty(
            IObjectWidgetSingleSubIntegration['single_textline'])
        single_date = FieldProperty(IObjectWidgetSingleSubIntegration['single_date'])
        single_readonly = FieldProperty(
            IObjectWidgetSingleSubIntegration['single_readonly'])


    class IObjectWidgetSingleIntegration(Interface):
        """Single integration interface"""
        subobj = Object(title=OBJECT_LABEL,
                        schema=IObjectWidgetSingleSubIntegration)


    @implementer(IObjectWidgetSingleIntegration)
    class ObjectWidgetSingleIntegration:
        """Single integration class"""
        subobj = FieldProperty(IObjectWidgetSingleIntegration['subobj'])


    class IObjectWidgetMultiSubIntegration(Interface):
        """Multi sub-integration interface"""
        multi_int = Int(title=INT_LABEL)
        multi_bool = Bool(title=BOOL_LABEL)
        multi_choice = Choice(title=CHOICE_LABEL,
                              values=('one', 'two', 'three'))
        multi_choice_opt = Choice(title=CHOICE_OPT_LABEL,
                                  values=('four', 'five', 'six'),
                                  required=False)
        multi_textline = TextLine(title=TEXTLINE_LABEL)
        multi_date = Date(title=DATE_LABEL)


    @implementer(IObjectWidgetMultiSubIntegration)
    class ObjectWidgetMultiSubIntegration(IntegrationBase):
        """Multi sub-integration class"""
        multi_int = FieldProperty(IObjectWidgetMultiSubIntegration['multi_int'])
        multi_bool = FieldProperty(IObjectWidgetMultiSubIntegration['multi_bool'])
        multi_choice = FieldProperty(IObjectWidgetMultiSubIntegration['multi_choice'])
        multi_choice_opt = FieldProperty(
            IObjectWidgetMultiSubIntegration['multi_choice_opt'])
        multi_textline = FieldProperty(
            IObjectWidgetMultiSubIntegration['multi_textline'])
        multi_date = FieldProperty(IObjectWidgetMultiSubIntegration['multi_date'])


    class IObjectWidgetMultiIntegration(Interface):
        """Multi-integration interface"""
        subobj = Object(title=OBJECT_LABEL,
                        schema=IObjectWidgetMultiSubIntegration)


    @implementer(IObjectWidgetMultiIntegration)
    class ObjectWidgetMultiIntegration:
        """Multi-integration class"""
        subobj = FieldProperty(IObjectWidgetMultiIntegration['subobj'])


    class IMultiWidgetListIntegration(Interface):
        """Multi-list interface"""
        list_of_int = List(title="ListOfInt label",
                           value_type=Int(title=INT_LABEL))
        list_of_bool = List(title="ListOfBool label",
                            value_type=Bool(title=BOOL_LABEL))
        list_of_choice = List(title="ListOfChoice label",
                              value_type=Choice(title=CHOICE_LABEL,
                                                values=('one', 'two', 'three')))
        list_of_textline = List(title="ListOfTextLine label",
                                value_type=TextLine(title=TEXTLINE_LABEL), )
        list_of_date = List(title="ListOfDate label",
                            value_type=Date(title=DATE_LABEL))
        list_of_objects = List(title="ListOfObject label",
                               value_type=Object(title=OBJECT_LABEL,
                                                 schema=IObjectWidgetMultiSubIntegration))


    @implementer(IMultiWidgetListIntegration)
    class MultiWidgetListIntegration(IntegrationBase):
        """Mulit-list class"""
        list_of_int = FieldProperty(IMultiWidgetListIntegration['list_of_int'])
        list_of_bool = FieldProperty(IMultiWidgetListIntegration['list_of_bool'])
        list_of_choice = FieldProperty(IMultiWidgetListIntegration['list_of_choice'])
        list_of_textline = FieldProperty(IMultiWidgetListIntegration['list_of_textline'])
        list_of_date = FieldProperty(IMultiWidgetListIntegration['list_of_date'])
        list_of_objects = FieldProperty(IMultiWidgetListIntegration['list_of_objects'])


    class IMultiWidgetDictIntegration(Interface):
        """Multi-dict interface"""
        dict_of_int = Dict(title="DictOfInt label",
                           key_type=Int(title='Int key'),
                           value_type=Int(title=INT_LABEL))
        dict_of_bool = Dict(title="DictOfBool label",
                            key_type=Bool(title='Bool key'),
                            value_type=Bool(title=BOOL_LABEL))
        dict_of_choice = Dict(title="DictOfChoice label",
                              key_type=Choice(title='Choice key',
                                              values=('key1', 'key2', 'key3')),
                              value_type=Choice(title=CHOICE_LABEL,
                                                values=('one', 'two', 'three')))
        dict_of_textline = Dict(title="DictOfTextLine label",
                                key_type=TextLine(title='TextLine key'),
                                value_type=TextLine(title=TEXTLINE_LABEL))
        dict_of_date = Dict(title="DictOfDate label",
                            key_type=Date(title='Date key'),
                            value_type=Date(title=DATE_LABEL))
        dict_of_objects = Dict(title="DictOfObject label",
                               key_type=TextLine(title='Object key'),
                               value_type=Object(title=OBJECT_LABEL,
                                                 schema=IObjectWidgetMultiSubIntegration))


    @implementer(IMultiWidgetDictIntegration)
    class MultiWidgetDictIntegration(IntegrationBase):
        """Multi-dict class"""
        dict_of_int = FieldProperty(IMultiWidgetDictIntegration['dict_of_int'])
        dict_of_bool = FieldProperty(IMultiWidgetDictIntegration['dict_of_bool'])
        dict_of_choice = FieldProperty(IMultiWidgetDictIntegration['dict_of_choice'])
        dict_of_textline = FieldProperty(IMultiWidgetDictIntegration['dict_of_textline'])
        dict_of_date = FieldProperty(IMultiWidgetDictIntegration['dict_of_date'])
        dict_of_objects = FieldProperty(IMultiWidgetDictIntegration['dict_of_objects'])


    def setup_form_defaults(registry):
        """Setup form default settings"""
        # Generic utilities
        registry.registerUtility(PyAMSSkin, provided=ISkin, name=PYAMS_BASE_SKIN_NAME)


    class IAJAXTestContent(Interface):
        """AJAX test content interface"""
        name = TextLine(title="Name")
        value = Int(title="Value")


    @implementer(IAJAXTestContent)
    class AJAXTestContent:
        """AJAX test content"""
        name = FieldProperty(IAJAXTestContent['name'])
        value = FieldProperty(IAJAXTestContent['value'])

        def __init__(self, **data):
            self.name = data.get('name')
            self.value = data.get('value')


    class ITestAddForm(IAJAXForm):
        """Test add form marker interface"""


    @ajax_form_config(name='test-add-form.html',
                      ajax_implements=ITestAddForm, ajax_method=None, ajax_xhr=None)
    class TestAddForm(AddForm):
        """Test add form"""
        fields = field.Fields(IAJAXTestContent)

        def create(self, data):
            return AJAXTestContent(**data)

        def add(self, obj):  # pylint: disable=redefined-builtin
            self.context[obj.name.lower()] = obj


    class ITestEditForm(IAJAXForm):
        """Test edit form marker interface"""


    @ajax_form_config(name='test-edit-form.html', context=IAJAXTestContent,
                      ajax_implements=ITestEditForm, ajax_method=None, ajax_xhr=None)
    class TestEditForm(EditForm):
        """Test edit form"""
        fields = field.Fields(IAJAXTestContent)



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.util module

Forms management utilities.
"""

import re
from binascii import hexlify
from collections import OrderedDict
from functools import total_ordering
from string import ascii_letters, digits

from six import class_types
from zope.contenttype import guess_content_type
from zope.interface import alsoProvides, directlyProvidedBy, implementer
from zope.interface.interface import InterfaceClass
from zope.interface.interfaces import ISpecification
from zope.schema.interfaces import IObject

from pyams_form.interfaces import IManager, ISelectionManager
from pyams_form.interfaces.form import IContextAware
from pyams_utils.interfaces.form import IDataManager
from pyams_utils.registry import get_current_registry


__docformat__ = 'restructuredtext'

from pyams_form import _  # pylint: disable=ungrouped-imports


_IDENTIFIER = re.compile('[A-Za-z][a-zA-Z0-9_]*$')
_ACCEPTABLE_CHARS = ascii_letters + digits + '_-'


def to_unicode(obj):
    """Convert object to unicode"""
    if isinstance(obj, bytes):
        return obj.decode('utf-8', 'ignore')
    return str(obj)


def to_bytes(obj):
    """Convert object to bytes"""
    if isinstance(obj, bytes):
        return obj
    if isinstance(obj, str):
        return obj.encode('utf-8')
    return str(obj).encode('utf-8')


def create_id(name):
    """Returns a *native* string as id of the given name."""
    if _IDENTIFIER.match(name):
        return str(name).lower()
    return hexlify(name.encode('utf-8')).decode()


@total_ordering
class MinType:
    """Minimum value"""

    def __le__(self, other):
        return True

    def __eq__(self, other):
        return self is other


def sorted_none(items):
    """Items sorter"""
    min_type = MinType()
    return sorted(items, key=lambda x: min_type if x is None else x)


def create_css_id(name):
    """Create CSS id"""
    return str(''.join([
        (char if char in _ACCEPTABLE_CHARS else
         hexlify(char.encode('utf-8')).decode())
        for char in name]))


def get_specification(spec, force=False):
    """Get the specification of the given object.

    If the given object is already a specification acceptable to the component
    architecture, it is simply returned. This is true for classes
    and specification objects (which includes interfaces).

    In case of instances, an interface is generated on the fly and tagged onto
    the object. Then the interface is returned as the specification.
    """
    # If the specification is an instance, then we do some magic.
    if (force or
            (spec is not None and
             not ISpecification.providedBy(spec) and
             not isinstance(spec, class_types))):

        # Step 1: Calculate an interface name
        iface_name = 'IGeneratedForObject_%i' % id(spec)

        # Step 2: Find out if we already have such an interface
        existing_interfaces = [
            i for i in directlyProvidedBy(spec) if i.__name__ == iface_name
        ]

        # Step 3a: Return an existing interface if there is one
        if len(existing_interfaces) > 0:
            spec = existing_interfaces[0]
        # Step 3b: Create a new interface if not
        else:
            iface = InterfaceClass(iface_name)
            alsoProvides(spec, iface)
            spec = iface
    return spec


def expand_prefix(prefix):
    """Expand prefix string by adding a trailing period if needed.

    expandPrefix(p) should be used instead of p+'.' in most contexts.
    """
    if prefix and not prefix.endswith('.'):
        return prefix + '.'
    return prefix


def get_widget_by_id(form, widget_id):
    """Get a widget by it's rendered DOM element id."""
    # convert the id to a name
    name = widget_id.replace('-', '.')
    prefix = form.prefix + form.widgets.prefix
    if not name.startswith(prefix):
        raise ValueError("Name %r must start with prefix %r" % (name, prefix))
    short_name = name[len(prefix):]
    return form.widgets.get(short_name, None)


def extract_content_type(form, widget_id):
    """Extract the content type of the widget with the given id."""
    widget = get_widget_by_id(form, widget_id)
    return guess_content_type(widget.filename)[0]


def extract_file_name(form, widget_id, cleanup=True, allow_empty_postfix=False):
    """Extract the filename of the widget with the given id.

    Uploads from win/IE need some cleanup because the filename includes also
    the path. The option ``cleanup=True`` will do this for you. The option
    ``allowEmptyPostfix`` allows to have a filename without extensions. By
    default this option is set to ``False`` and will raise a ``ValueError`` if
    a filename doesn't contain a extension.
    """
    widget = get_widget_by_id(form, widget_id)
    clean_file_name = ''
    dotted_parts = []
    if not allow_empty_postfix or cleanup:
        # We need to strip out the path section even if we do not reomve them
        # later, because we just need to check the filename extension.
        clean_file_name = widget.filename.split('\\')[-1]
        clean_file_name = clean_file_name.split('/')[-1]
        dotted_parts = clean_file_name.split('.')
    if not allow_empty_postfix:
        if len(dotted_parts) <= 1:
            raise ValueError(_('Missing filename extension.'))
    if cleanup:
        return clean_file_name
    return widget.filename


def changed_field(field, value, context=None):
    """Figure if a field's value changed

    Comparing the value of the context attribute and the given value"""
    if context is None:
        context = field.context
    if context is None:
        # IObjectWidget madness
        return True
    if IObject.providedBy(field):
        return True

    # Get the datamanager and get the original value
    dman = get_current_registry().getMultiAdapter((context, field), IDataManager)
    # now figure value changed status
    # Or we can not get the original value, in which case we can not check
    # Or it is an Object, in case we'll never know
    if (not dman.can_access()) or (dman.query() != value):
        return True
    return False


def changed_widget(widget, value, field=None, context=None):
    """Figure if a widget's value changed

    Comparing the value of the widget context attribute and the given value.
    """
    if IContextAware.providedBy(widget) and not widget.ignore_context:
        # if the widget is context aware, figure if it's field changed
        if field is None:
            field = widget.field
        if context is None:
            context = widget.context
        return changed_field(field, value, context=context)
    # otherwise we cannot, return 'always changed'
    return True


@implementer(IManager)
class Manager(OrderedDict):
    """Non-persistent IManager implementation."""

    def create_according_to_list(self, data, order):
        """Arrange elements from data according to sorting of given list"""
        self.clear()
        for key in order:
            if key in data:
                self[key] = data[key]

    def __getitem__(self, key):
        if key not in self:
            try:
                return getattr(self, key)
            except AttributeError:
                # make sure a KeyError is raised later
                pass
        return super().__getitem__(key)


@implementer(ISelectionManager)
class SelectionManager(Manager):
    """Non-persistent ISelectionManager implementation."""

    manager_interface = None

    def __add__(self, other):
        if not self.manager_interface.providedBy(other):
            return NotImplemented
        return self.__class__(self, other)

    def select(self, *names):
        """See interfaces.ISelectionManager"""
        return self.__class__(*[self[name] for name in names])

    def omit(self, *names):
        """See interfaces.ISelectionManager"""
        return self.__class__(
            *[item for name, item in self.items()
              if name not in names])

    def copy(self):
        """See interfaces.ISelectionManager"""
        return self.__class__(*self.values())



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.action module

This module defines all actions.
"""

from zope.interface import implementer, implementer_only

from pyams_form.interfaces.button import ActionExecutionError, IAction, IActionErrorEvent, \
    IActionEvent, IActionHandler, IActions
from pyams_form.util import Manager, create_id


__docformat__ = 'restructuredtext'


@implementer(IActionEvent)
class ActionEvent:
    """Base action event class"""

    def __init__(self, action):
        self.action = action

    def __repr__(self):
        return '<%s for %r>' % (self.__class__.__name__, self.action)


@implementer(IActionErrorEvent)
class ActionErrorOccurred(ActionEvent):
    """An event telling the system that an error occurred during action
    execution."""

    def __init__(self, action, error):
        super().__init__(action)
        self.error = error


class ActionSuccessful(ActionEvent):
    """An event signalizing that an action has been successfully executed."""


@implementer(IAction)
class Action:
    """Action class."""

    __name__ = __parent__ = None

    def __init__(self, request, title, name=None):
        self.request = request
        self.title = title
        if name is None:
            name = create_id(title)
        self.name = name

    def is_executed(self):
        """Check if action was executed by looking for action name into request params"""
        return self.name in self.request.params

    def __repr__(self):
        return '<%s %r %r>' % (self.__class__.__name__, self.name, self.title)


@implementer_only(IActions)
class Actions(Manager):
    """Action manager class."""

    __name__ = __parent__ = None

    def __init__(self, form, request, content):
        super().__init__()
        self.form = form
        self.request = request
        self.content = content

    @property
    def executed_actions(self):
        """Get list of executed actions"""
        return [action for action in self.values()
                if action.is_executed()]

    def update(self):
        """See pyams_form.interfaces.button.IActions."""

    def execute(self):
        """See pyams_form.interfaces.button.IActions."""
        registry = self.request.registry
        for action in self.executed_actions:
            handler = registry.queryMultiAdapter((self.form, self.request, self.content, action),
                                                 IActionHandler)
            if handler is not None:
                try:
                    result = handler()
                except ActionExecutionError as error:
                    registry.notify(ActionErrorOccurred(action, error))
                else:
                    registry.notify(ActionSuccessful(action))
                    return result
        return None

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.__name__)


@implementer(IActionHandler)
class ActionHandlerBase:
    """Action handler base adapter."""

    def __init__(self, form, request, content, action):
        self.form = form
        self.request = request
        self.content = content
        self.action = action



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.textlines module

This module provides default textlines widget.
"""

from zope.interface import implementer, implementer_only

from pyams_form.browser.textarea import TextAreaWidget
from pyams_form.interfaces.widget import IFieldWidget, ITextLinesWidget
from pyams_form.widget import FieldWidget
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config
from pyams_utils.schema import ITextLineListField


__docformat__ = 'restructuredtext'


@implementer_only(ITextLinesWidget)
class TextLinesWidget(TextAreaWidget):
    """Input type sequence widget implementation."""


@adapter_config(required=(ITextLineListField, IFormLayer),
                provides=IFieldWidget)
def TextLinesFieldWidget(field, request):  # pylint: disable=invalid-name
    """IFieldWidget factory for TextLinesWidget."""
    return FieldWidget(field, TextLinesWidget(request))


@implementer(IFieldWidget)
def TextLinesFieldWidgetFactory(field, value_type, request):
    # pylint: disable=invalid-name,unused-argument
    """IFieldWidget factory for TextLinesWidget."""
    return TextLinesFieldWidget(field, request)



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.interfaces module

This module defines all package interfaces.
"""

from zope.interface import Interface
from zope.interface.common.mapping import IEnumerableMapping
from zope.schema import Bool, Field, TextLine


__docformat__ = 'reStructuredText'

from pyams_form import _


# Forms modes

INPUT_MODE = 'input'
DISPLAY_MODE = 'display'
HIDDEN_MODE = 'hidden'


#
# Generic manager interfaces
#

class IManager(IEnumerableMapping):
    """A manager of some kind of items.

    *Important*: While managers are mappings, the order of the items is
    assumed to be important! Effectively, a manager is an ordered mapping.

    In general, managers do not have to support a manipulation
    API. Oftentimes, managers are populated during initialization or while
    updating.
    """


class ISelectionManager(IManager):
    """Managers that support item selection and management.

    This manager allows one to more carefully specify the contained items.

    *Important*: The API is chosen in a way, that the manager is still
    immutable. All methods in this interface must return *new* instances of
    the manager.
    """

    def __add__(self, other):
        """Used for merge two managers."""

    def select(self, *names):
        """Return a modified instance with an ordered subset of items."""

    def omit(self, *names):
        """Return a modified instance omitting given items."""

    def copy(self):
        """Copy all items to a new instance and return it."""


#
# Forms validators
#

class IData(Interface):
    """A proxy object for form data.

    The object will make all keys within its data attribute available as
    attributes. The schema that is represented by the data will be directly
    provided by instances.
    """

    def __init__(self, schema, data, context):  # pylint: disable=super-init-not-called
        """The data proxy is instantiated using the schema it represents, the
        data fulfilling the schema and the context in which the data are
        validated.
        """

    __context__ = Field(title=_('Context'),
                        description=_('The context in which the data are validated'),
                        required=True)


class IValidator(Interface):
    """A validator for a particular value"""

    def validate(self, value, force=False):
        """Validate the value.

        If successful, return ``None``. Otherwise raise an ``Invalid`` error.
        """


class IManagerValidator(Interface):
    """A validator that validates a set of data."""

    def validate(self, data):
        """Validate a dictionary of data.

        This method is only responsible of validating relationships between
        the values in the data. It can be assumed that all values have been
        validated in isolation before.

        The return value of this method is a tuple of errors that occurred
        during the validation process.
        """

    def validate_object(self, obj):
        """Validate an object.

        The same semantics as in ``validate()`` apply, except that the values
        are retrieved from the object and not the data dictionary.
        """


#
# Forms fields
#

class IField(Interface):
    """Field wrapping a schema field used in the form."""

    __name__ = TextLine(title=_('Title'),
                        description=_('The name of the field within the form'),
                        required=True)

    field = Field(title=_('Schema Field'),
                  description=_('The schema field that is to be rendered'),
                  required=True)

    prefix = Field(title=_('Prefix'),
                   description=_('The prefix of the field used to avoid name clashes'),
                   required=True)

    mode = Field(title=_('Mode'),
                 description=_('The mode in which to render the widget for the field'),
                 required=True)

    interface = Field(title=_('Interface'),
                      description=_('The interface from which the field is coming'),
                      required=True)

    ignore_context = Bool(title=_('Ignore Context'),
                          description=_('A flag, when set, forces the widget not to look at '
                                        'the context for a value'),
                          required=False)

    widget_factory = Field(title=_('Widget Factory'),
                           description=_('The widget factory'),
                           required=False,
                           default=None,
                           missing_value=None)

    show_default = Bool(title=_('Show default value'),
                        description=_('A flag, when set, makes the widget to display '
                                      'field|adapter provided default values'),
                        default=True,
                        required=False)


class IFields(ISelectionManager):
    """IField manager."""

    def select(self, *names, prefix=None, interface=None):  # pylint: disable=arguments-differ
        """Return a modified instance with an ordered subset of items.

        This extension to the ``ISelectionManager`` allows for handling cases
        with name-conflicts better by separating field selection and prefix
        specification.
        """

    def omit(self, *names, prefix=None, interface=None):  # pylint: disable=arguments-differ
        """Return a modified instance omitting given items.

        This extension to the ``ISelectionManager`` allows for handling cases
        with name-conflicts better by separating field selection and prefix
        specification.
        """


class IContentProviders(IManager):
    """A content provider manager"""


#
# Forms data converters
#

class IDataConverter(Interface):
    """A data converter from field to widget values and vice versa"""

    def to_widget_value(self, value):
        """Convert the field value to a widget output value.

        If conversion fails or is not possible, a ``ValueError`` *must* be
        raised. However, this method should effectively never fail, because
        incoming value is well-defined.
        """

    def to_field_value(self, value):
        """Convert an input value to a field/system internal value.

        This methods *must* also validate the converted value against the
        field.

        If the conversion fails, a ``ValueError`` *must* be raised. If
        the validation fails, a ``ValidationError`` *must* be raised.
        """


#
# Widgets and fields values
#

class IValue(Interface):
    """A value"""

    def get(self):  # pylint: disable=arguments-differ
        """Returns the value."""


#
# Terms interfaces
#

class IVocabularyTerms(Interface):
    """Base vocabulary terms lookup interface

    We keep camelCase naming convention here to be able to use zope.schema vocabularies
    easilly!
    """

    def getTerm(self, value):  # pylint: disable=invalid-name
        """Return an ITitledTokenizedTerm object for the given value

        LookupError is raised if the value isn't in the source
        """

    def getValue(self, token):  # pylint: disable=invalid-name
        """Return a value for a given identifier token

        LookupError is raised if there isn't a value in the source.
        """


class ITerms(IVocabularyTerms):
    """A selection term"""

    context = Field()
    request = Field()
    form = Field()
    field = Field()
    widget = Field()

    def getTermByToken(self, token):  # pylint: disable=invalid-name
        """Return an ITokenizedTerm for the passed-in token.

        If `token` is not represented in the vocabulary, `LookupError`
        is raised.
        """

    def __iter__(self):
        """Iterate over terms."""

    def __len__(self, ):
        """Return number of terms."""

    def __contains__(self, value):
        """Check wether terms containes the ``value``."""


class IBoolTerms(ITerms):
    """A specialization that handles boolean choices."""

    true_label = TextLine(title=_('True-value Label'),
                          description=_('The label for a true value of the Bool field'),
                          required=True)

    false_label = TextLine(title=_('False-value Label'),
                           description=_('The label for a false value of the Bool field'),
                           required=False)


#
# Objects factories
#

class IObjectFactory(Interface):
    """Factory that will instantiate our objects for ObjectWidget.

    It could also pre-populate properties as it gets the values extracted
    from the form passed in ``value``.
    """

    def __call__(self, value):  # pylint: disable=signature-differs
        """Return a default object created to be populated.
        """


#
# Subform factories
#

class ISubformFactory(Interface):
    """Factory that will instantiate our subforms for ObjectWidget"""

    def __call__(self):  # pylint: disable=signature-differs
        """Return a default object created to be populated"""


#
# Widgets layout template
# This interface isn't into "widget" module to avoid cyclic dependencies...
#

class IWidgetLayoutTemplate(Interface):
    """Widget layout template marker used for render the widget layout.

    It is important that we don't inherit this template from IPageTemplate.
    otherwise we will get into trouble since we lookup an IPageTemplate
    in the widget/render method.
    """



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.submit module

This module provides form submit button widget.
"""

from zope.interface import implementer_only

from pyams_form.browser.button import ButtonWidget
from pyams_form.interfaces.button import IButton
from pyams_form.interfaces.widget import ISubmitWidget, IFieldWidget
from pyams_form.widget import FieldWidget
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config


__docformat__ = 'restructuredtext'


@implementer_only(ISubmitWidget)
class SubmitWidget(ButtonWidget):
    """Form submit button."""

    klass = 'submit-widget'
    css = 'submit'

    def json_data(self):
        data = super().json_data()
        data['type'] = 'submit'
        return data


@adapter_config(required=(IButton, IFormLayer), provides=IFieldWidget)
def SubmitFieldWidget(field, request):  # pylint: disable=invalid-name
    """Form submit button factory adapter"""
    submit = FieldWidget(field, SubmitWidget(request))
    submit.value = field.title
    return submit
